#include "TM4C123GH6PM.h"
#include <stdlib.h>



// Define the main system clock frequency
#define SYS_CLK 50e6

// Define bit positions for easy manipulation
#define D0   (1U << 0)  // Bit 0
#define D1   (1U << 1)  // Bit 1
#define D2   (1U << 2)  // Bit 2
#define D3   (1U << 3)  // Bit 3
#define D4   (1U << 4)  // Bit 4
#define D5   (1U << 5)  // Bit 5
#define D6   (1U << 6)  // Bit 6
#define D7   (1U << 7)  // Bit 7

// Define port aliases for readability
#define PORTA D0
#define PORTC D2
#define PORTD D3

// Define pin aliases for GPIO operations
#define PIN2 D2
#define PIN3 D3
#define PIN4 D4
#define PIN5 D5
#define PIN6 D6
#define PIN7 D7

// Define output pins for LEDs
#define YELLOW_OUTPUT_PIN PIN4 // LED on PA4
#define RED_OUTPUT_PIN    PIN5 // LED on PA5
#define BLUE_OUTPUT_PIN   PIN6 // LED on PA6
#define WHITE_OUTPUT_PIN  PIN7 // LED on PA7

// Define input pins for buttons
#define YELLOW_INPUT_PIN PIN4  // Button on PC4
#define RED_INPUT_PIN    PIN5  // Button on PC5
#define BLUE_INPUT_PIN   PIN6  // Button on PC6
#define WHITE_INPUT_PIN  PIN7  // Button on PC7

// Game configurations
#define ARRAY_SIZE 32

// LED color symbols
#define YELLOW  'Y'
#define RED     'R'
#define WHITE   'W'
#define BLUE    'B'

// Frequencies corresponding to the notes for LEDs
#define YELLOW_FREQ 262  // Note for Yellow LED
#define RED_FREQ    330  // Note for Red LED
#define WHITE_FREQ  523  // Note for White LED
#define BLUE_FREQ   494  // Note for Blue LED

// Array of tones for the "Game Over" sound sequence
const float tones[] = {349.23, 392.00, 440.00, 493.88, 523.25};

// Global variables for game state
int usr_level = 1;             // Current game level
int usr_input_counter = 0;     // Tracks user's input sequence
char triggered_pin;            // Current user-triggered button
char randomArray[ARRAY_SIZE];  // Randomized game sequence

// Function declarations
void generateRandomArray(char *array);
void playPattern(char *array);
void playLED(char led);
void playSound(int frequency);
void configureWTimer1AforDelays(void);
void delaySec(uint16_t sec);
void delayMicroSec(int sec);
void shutSpeaker(void);
void configureGPIOA_PA4toPA7(void);
void configureGPIOC_PC4toPC7(void);
void configureWTimer3APWM(void);
void playGameOver(void);
void getPressedButton(uint32_t gpioIntStatus);
void ADC0_init();
int analogRead();

int main() {
    // Initialize GPIO for LEDs and buttons
    configureGPIOA_PA4toPA7();
    configureGPIOC_PC4toPC7();
    
    // Configure timers for delays and PWM control
    configureWTimer1AforDelays();
    configureWTimer3APWM();
    
    // Initialize ADC for random seed generation
    ADC0_init();
    srand(analogRead());  // Seed the random generator with ADC reading

    // Generate the initial random game sequence
    generateRandomArray(randomArray);
    
    // Play the pattern corresponding to the user's current level
    playPattern(randomArray);
    
    // Enable global interrupts
    __enable_irq();
    
    // Main loop to run the game indefinitely
    while (1) {
        // Wait for user interaction or game logic updates
    }
}

/**
 * Generate a random sequence of colors ('Y', 'W', 'R', 'B') for the game.
 * @param array Pointer to the array that will store the generated sequence.
 * The sequence length is defined by ARRAY_SIZE.
 */

void generateRandomArray(char *array) {
    // Generate a sequence of random colors
    for (int i = 0; i < ARRAY_SIZE; i++) {
        array[i] = "YWRB"[rand() % 4];
    }
}

/**
 * Initialize ADC0 and configure Sequencer 3 for analog input on PD3.
 * Configures the GPIO pin PD3 for analog mode and sets up ADC0 for
 * single-ended input using software triggering.
 */

void ADC0_init() {
    // Enable the clock for ADC0 module
    SYSCTL->RCGCADC |= D0;

    // Short delay to allow the clock to stabilize
    for (volatile int i = 0; i < 3; i++);

    // Enable the clock for GPIO Port D
    SYSCTL->RCGCGPIO |= PORTD;

    // Enable alternate function for pin PD3
    GPIOD->AFSEL |= PIN3;

    // Disable the digital functionality on PD3 (PD3 will function as an analog pin)
    GPIOD->DEN &= ~PIN3;

    // Enable analog mode on PD3
    GPIOD->AMSEL |= PIN3;

    // Disable ADC Sequencer 3 (SS3) during configuration
    ADC0->ACTSS &= ~D3;

    // Set the ADC sequencer to software trigger mode by clearing the EMUX bits
    ADC0->EMUX &= ~0xF000;

    // Configure SS3 to use input channel 4 (PD3 is connected to ADC input channel 4)
    ADC0->SSMUX3 = 4;

    // Configure SS3 to generate an interrupt on completion and mark the last sample
    ADC0->SSCTL3 = 0x06;

    // Re-enable ADC Sequencer 3 (SS3) after configuration
    ADC0->ACTSS |= D3;
}

/**
 * Perform an analog-to-digital conversion using ADC0 Sequencer 3.
 * @return The converted analog value as a integer.
 */
int analogRead() {
    // Start the sampling and conversion process on ADC Sequencer 3
    ADC0->PSSI |= D3;

    // Wait for the conversion to complete (polling the RIS register)
    while (!(ADC0->RIS & D3));

    // Retrieve the ADC conversion result from the Sequencer 3 FIFO
    int adcResult = ADC0->SSFIFO3;

    // Clear the interrupt flag for Sequencer 3 to prepare for the next conversion
    ADC0->ISC |= D3;

    // Return the ADC conversion result
    return adcResult;
}



/**
 * Play the current game pattern up to the user's level
 * @param array Pointer to the array containing the game sequence
 */
void playPattern(char *array) {
    for (int i = 0; i < usr_level; i++) {
        playLED(array[i]);
    }
}

/**
 * Play LED and sound for a specific color
 * @param led Color to play (YELLOW, WHITE, BLUE, or RED)
 */
void playLED(char led) {
    // Determine which LED to light up and play the corresponding sound
    switch (led) {
        case YELLOW:
            // Turn on the Yellow LED (connected to PA4) and play the Yellow tone
            GPIOA->DATA |= YELLOW_OUTPUT_PIN;
            playSound(YELLOW_FREQ);
            break;

        case WHITE:
            // Turn on the White LED (connected to PA7) and play the White tone
            GPIOA->DATA |= WHITE_OUTPUT_PIN;
            playSound(WHITE_FREQ);
            break;

        case BLUE:
            // Turn on the Blue LED (connected to PA6) and play the Blue tone
            GPIOA->DATA |= BLUE_OUTPUT_PIN;
            playSound(BLUE_FREQ);
            break;

        case RED:
            // Turn on the Red LED (connected to PA5) and play the Red tone
            GPIOA->DATA |= RED_OUTPUT_PIN;
            playSound(RED_FREQ);
            break;
    }

    // Wait for 300 microseconds to keep the LED and sound on
    delayMicroSec(300);

    // Turn off all LEDs (clear bits PA4–PA7) and the speaker
    GPIOA->DATA &= ~0xF0; 
    shutSpeaker();
		
		
    delayMicroSec(400);

    
}


/**
 * Configure Wide Timer 1 for delay operations
 */
void configureWTimer1AforDelays() {
    // Enable the clock for Wide Timer 1
    SYSCTL->RCGCWTIMER |= D1;

    // Ensure Timer A is disabled before configuring it
    WTIMER1->CTL &= ~D0;

    // Set the timer to a 32-bit configuration (individual timer mode)
    WTIMER1->CFG = 0x00;

    // Configure Timer A for one-shot mode (TAMR = 0x1)
    WTIMER1->TAMR = 0x01;

    // Clear any pending Timer A interrupts by writing to the ICR register
    WTIMER1->ICR = D0;
}


/**
 * Delay for specified number of seconds
 * @param sec Number of seconds to delay
 */
void delaySec(uint16_t sec) {
    WTIMER1->TAILR = 50000000 - 1;
    for (uint16_t i = 0; i < sec; i++) {
        WTIMER1->CTL |= D0;
        while (!(WTIMER1->RIS & D0));
        WTIMER1->ICR = D0;
    }
}

/**
 * Delay for specified number of microseconds
 * @param sec Number of microseconds to delay
 */
void delayMicroSec(int sec) {
    WTIMER1->TAILR = 50000 - 1;
    for (int i = 0; i < sec; i++) {
        WTIMER1->CTL |= D0;
        while (!(WTIMER1->RIS & D0));
        WTIMER1->ICR = D0;
    }
}

/**
 * Configure GPIO Port A for LED outputs
 */
void configureGPIOA_PA4toPA7(void) {
    // Enable the clock for GPIO Port A
    SYSCTL->RCGCGPIO |= PORTA;

    // Set PA4 to PA7 as output pins (Yellow, Red, White, Blue LEDs)
    GPIOA->DIR |= (YELLOW_OUTPUT_PIN | RED_OUTPUT_PIN | WHITE_OUTPUT_PIN | BLUE_OUTPUT_PIN);

    // Enable digital functionality for PA4 to PA7
    GPIOA->DEN |= (YELLOW_OUTPUT_PIN | RED_OUTPUT_PIN | WHITE_OUTPUT_PIN | BLUE_OUTPUT_PIN);
}


/**
 * Configure GPIO Port C for button inputs with interrupts
 */
void configureGPIOC_PC4toPC7(void) {
    // Enable the clock for GPIO Port C
    SYSCTL->RCGCGPIO |= PORTC;

    // Set PC4 to PC7 as input pins (clear corresponding DIR bits)
    GPIOC->DIR &= ~(YELLOW_INPUT_PIN | RED_INPUT_PIN | WHITE_INPUT_PIN | BLUE_INPUT_PIN);

    // Enable digital functionality for PC4 to PC7
    GPIOC->DEN |= (YELLOW_INPUT_PIN | RED_INPUT_PIN | WHITE_INPUT_PIN | BLUE_INPUT_PIN);

    // Enable pull-down resistors for PC4 to PC7 to ensure a defined state when unpressed
    GPIOC->PDR |= (YELLOW_INPUT_PIN | RED_INPUT_PIN | WHITE_INPUT_PIN | BLUE_INPUT_PIN);

    // Set PC4 to PC7 for edge-triggered interrupts (clear IS bits for edge sensitivity)
    GPIOC->IS &= ~(YELLOW_INPUT_PIN | RED_INPUT_PIN | WHITE_INPUT_PIN | BLUE_INPUT_PIN);

    // Configure PC4 to PC7 for single edge detection (clear IBE bits for single edge)
    GPIOC->IBE &= ~(YELLOW_INPUT_PIN | RED_INPUT_PIN | WHITE_INPUT_PIN | BLUE_INPUT_PIN);

    // Set PC4 to PC7 to trigger on rising edges (set IEV bits)
    GPIOC->IEV |= (YELLOW_INPUT_PIN | RED_INPUT_PIN | WHITE_INPUT_PIN | BLUE_INPUT_PIN);

    // Enable interrupts for PC4 to PC7
    GPIOC->IM |= (YELLOW_INPUT_PIN | RED_INPUT_PIN | WHITE_INPUT_PIN | BLUE_INPUT_PIN);

    // Clear any prior interrupt flags for PC4 to PC7
    GPIOC->ICR |= (YELLOW_INPUT_PIN | RED_INPUT_PIN | WHITE_INPUT_PIN | BLUE_INPUT_PIN);

    // Enable GPIO Port C interrupts in the Nested Vectored Interrupt Controller (NVIC)
    NVIC_EnableIRQ(GPIOC_IRQn);
}

void GPIOC_Handler(void) {
    // Get a snapshot of the Masked Interrupt Status (MIS) register for Port C
    uint32_t gpioIntStatus = GPIOC->MIS;
    

    // Determine which button was pressed based on the interrupt status
    getPressedButton(gpioIntStatus);

    // Check if the button press matches the current sequence in the game
    if (triggered_pin == randomArray[usr_input_counter]) {
        // If correct, light the corresponding LED and play the sound
        playLED(triggered_pin);

        // Move to the next button in the sequence
        usr_input_counter++;

        // Check if the user has completed the current level
        if (usr_input_counter == usr_level) {
            // Advance to the next level and reset the user input counter
            usr_level++;
            usr_input_counter = 0;
						
						// Wait for 600 microseconds before the next pattern or action
						delayMicroSec(600);

            // Replay the updated pattern for the new level
            playPattern(randomArray);
        }
    } else {
        // If incorrect, trigger the "Game Over" sequence
        playGameOver();
    }

			
		// Clear all interrupts for Port C by writing to the Interrupt Clear (ICR) register
    GPIOC->ICR = 0xFF;
		
		// Read back the ICR register to ensure the clear operation is complete
    volatile int temp = GPIOC->ICR;
		
}


/**
 * Play a sound at a specific frequency
 * @param frequency Frequency of the sound to play
 */
void playSound(int frequency) {
    // Enable the alternate function for PD2 to configure it as a timer output
    GPIOD->AFSEL |= (1 << 2); 

    // Calculate the timer load value based on the desired frequency
    int counts = SYS_CLK / frequency;

    // Set the Timer A Interval Load Register to the calculated value
    WTIMER3->TAILR = counts - 1;

    // Set the Timer A Match Register to create a 50% duty cycle for PWM
    WTIMER3->TAMATCHR = counts / 2;

    // Enable Timer A to start the PWM signal
    WTIMER3->CTL |= (1 << 0);
}


/**
 * Shut off the speaker
 */
void shutSpeaker(void) {
    // Disable the Wide Timer 3A to stop PWM generation
    WTIMER3->CTL &= ~(1 << 0);

    // Disable the alternate function for PD2, reverting it to a general-purpose pin
    GPIOD->AFSEL &= ~(1 << 2);

    // Clear the PD2 output data to ensure no residual signal is sent
    GPIOD->DATA &= ~(1 << 2);
}


/**
 * Configure Wide Timer 3A for PWM output
 */
void configureWTimer3APWM(void) {
    // Enable clock for GPIO Port D
    SYSCTL->RCGCGPIO |= D3;

    // Enable clock for Wide Timer 3
    SYSCTL->RCGCWTIMER |= D3;

    // Enable digital function for PD2
    GPIOD->DEN |= PIN2;

    // Enable alternate function for PD2
    GPIOD->AFSEL |= PIN2;

    // Clear the PCTL bits for PD2 (configure as a general-purpose pin)
    GPIOD->PCTL &= ~0x00000F00;

    // Set PD2 as WT3CCP0 (Wide Timer 3 Capture/Compare/PWM output)
    GPIOD->PCTL |= 0x00000700;

    // Disable WTIMER3A for configuration
    WTIMER3->CTL &= ~D0;

    // Configure WTIMER3A as a 32-bit timer
    WTIMER3->CFG = 0x04;

    // Set WTIMER3A to PWM mode (TAMR: Periodic and PWM enabled)
    WTIMER3->TAMR = 0x0A;
}


/**
 * Get the pressed button from GPIO Port C
 * @param gpioIntStatus is a snapshot of Masked Interrupt register 
 */
void getPressedButton(uint32_t gpioIntStatus) {
    // Determine which button caused the interrupt by checking the status bits

    // Check if the Yellow button (connected to YELLOW_INPUT_PIN) triggered the interrupt
    if (gpioIntStatus & YELLOW_INPUT_PIN) {
        triggered_pin = 'Y'; // Set triggered_pin to 'Y' for Yellow
        return; // Exit the function
    }

    // Check if the Red button (connected to RED_INPUT_PIN) triggered the interrupt
    else if (gpioIntStatus & RED_INPUT_PIN) {
        triggered_pin = 'R'; // Set triggered_pin to 'R' for Red
        return; // Exit the function
    }

    // Check if the White button (connected to WHITE_INPUT_PIN) triggered the interrupt
    else if (gpioIntStatus & WHITE_INPUT_PIN) {
        triggered_pin = 'W'; // Set triggered_pin to 'W' for White
        return; // Exit the function
    }

    // Check if the Blue button (connected to BLUE_INPUT_PIN) triggered the interrupt
    else if (gpioIntStatus & BLUE_INPUT_PIN) {
        triggered_pin = 'B'; // Set triggered_pin to 'B' for Blue
        return; // Exit the function
    }

    // If no valid button press was detected, set triggered_pin to a null character
    triggered_pin = '\0';
}



/**
 * Play a sequence of tones and blinking LEDs to indicate a game over
 */
void playGameOver(void) {
    // Loop through the "Game Over" tones array
    for (uint8_t i = 0; i < sizeof(tones) / sizeof(tones[0]); i++) {
        // Play the current tone using the playSound function
        playSound(tones[i]);

        // Turn on all LEDs (PA4 to PA7) to indicate "Game Over"
        GPIOA->DATA |= 0xF0;

        // Keep the LEDs on for 200 microseconds
        delayMicroSec(200);

        // Turn off all LEDs (PA4 to PA7)
        GPIOA->DATA &= ~0xF0;

        // Wait for 200 microseconds before playing the next tone
        delayMicroSec(200);
    }

    // Shut off the speaker after the "Game Over" sequence
    shutSpeaker();
}
